from ...sources import Source
from ...indicators import Indicator
from ...pricing import Candle
from .side_pluckers import SidePlucker
from .candle_pluckers import CandlePlucker
from .row_pluckers import RowPlucker


def smart_plucker(source, side=Source.ASK, component='end', row=0):
    """
    Depending on the source's type, it will be
      smart enough to instantiate a plucker, or
      return the source as-is. In the end, it
      returns a mean to get a width=1 source.
    Also, if it is a source, it will use a specific
      side (and not both) to fetch.
    :param source: The source to wrap.
    :param side: The flow side to pick.
    :param component: The component to pluck.
      Used if the source is Candle-type.
    :param row: The row to pluck.
      Used if the source is an indicator.
    :return: A width=1 mapper.
    """

    if isinstance(source, Source):
        if source.dtype == Candle:
            return CandlePlucker(SidePlucker(source, side), component)
        elif source.dtype == int:
            return SidePlucker(source, side)
        else:
            raise TypeError("The source has an unexpected dtype")
    elif isinstance(source, Indicator):
        return RowPlucker(source, row)
    else:
        raise TypeError("The source must be either a Source or an Indicator")



class TailRunner:
    """
    A TailRunner works by appropriately fetching data to be iterated
      by overlapping chunks ("tails") ending in the desired (iterated)
      index, and of an -at most // when "complete"- size.
    """

    def __init__(self, tail_size):
        if not isinstance(tail_size, int) or tail_size <= 1:
            raise ValueError("Tail size must be an integer greater than 1")
        self._tail_size = tail_size

    @property
    def tail_size(self):
        return self._tail_size

    @staticmethod
    def _tail_slice(data, start, end, tail_size):
        """
        Creates a slice of the given data (which will most often a source of the
          same scale/interval of the current indicator, and often a dependency)
          by considering the start and end indices (which are often provided in
          the context of data update) and a positive integer tail size.

        The slice is created according to the following criteria:
          - First, the end index is usually as-is, since it is the pivot index
            to start calculating the tail, which extends backward.
          - The start index will be before the end index, and will be considered.
          - Then we subtract (tail_size - 1) to the start index: for each index
            iteration, we should consider the start of its tail is exactly
            (tail_size - 1) steps behind, and includes this step as the last.
          - However, when computing slices, it may occur that such new index is
            below 0, so we truncate it to 0. In such cases, the slicing will
            have less elements than following slices.

        An array is returned, which may actually have less elements than the
          result of (end - start + tail_size - 1), if the data has less elements.
        :param data: The data or source being sliced.
        :param start: The start index - usually in the context of an update operation.
        :param end: The end index - usually in the context of an update operation.
        :param tail_size: The tail size.
        :return: A slice of the given data.
        """

        return data[max(0, start + 1 - tail_size):end]

    @staticmethod
    def _tail_iterate(data, start, end, tail_size):
        """
        Teaming with the _tail_slice utility method, returns a generator which will
          iterate over the given data (which will most often be a dependency of the
          same scale), appropriately indexed to get the tail start, the tail end,
          whether a complete tail was fetched this iteration, and the actual slice
          of data corresponding to this iteration, given the overall start, end,
          and tail size for the given data.
        :param data: The data being iterated. It was already sliced by _tail_size.
        :param start: The start index - usually in the context of an update operation.
        :param end: The end index - usually in the context of an update operation.
        :param tail_size: The tail size.
        :return: A generator yielding tuples like:
          (local tail start index,
           local tail end index,
           tail is incomplete,
           global index)
        """

        offset = data.shape[0] - end + start
        for idx in range(0, end - start):
            tail_end = idx + 1 + offset
            tail_start = tail_end - tail_size
            incomplete = False
            if tail_start < 0:
                tail_start = 0
                incomplete = True
            yield tail_start, tail_end, incomplete, start + idx

    def tail_iterate(self, start, end, source):
        """
        Iterates over the [start:end] indices (not including the
          end) to get a tail, of the chosen tail size, ending in
          the index being iterated.
        :param start: The start index.
        :param end: The end index.
        :param source: The source of the iterator.
        :return: An iterator yielding (index, tail, incomplete)
          being the current end index, the tail, and a flag telling
          whether the tail is full-size or not (depending on the
          chosen index).
        """

        data = self._tail_slice(source, start, end, self._tail_size)
        for tail_start, tail_end, incomplete, idx in self._tail_iterate(data, start, end, self._tail_size):
            yield idx, data[tail_start:tail_end], incomplete



"""
Defined both the standardized price and the japanese candle. The japanese candle holds
  a range of values for a certain timelapse, stating its minimum, maximum, start and
  prices.

Prices are standardized to integer, but perhaps every price statistic has a different scaling
  mechanism. E.g. satoshis will not have scaling, but dollar/euro conversion will have 6
  digits of scaling for their floats. Those cases are converted to integers (e.g. by zero-pad
  and then point-removal).
"""


from collections import namedtuple
from numpy import uint64 as _uint64


StandardizedPrice = _uint64


class Candle:
    """
    Japanese candles span over intervals and keep a detail of start, end, min and max (standarized)
      prices of an asset in a (somehow neutral or reference) currency.
    """

    __slots__ = ('start', 'end', 'min', 'max')

    def __new__(cls, start, min, max, end):
        obj = super(Candle, cls).__new__(cls)
        obj.start = start
        obj.end = end
        obj.min = min
        obj.max = max
        return obj

    @classmethod
    def constant(cls, value):
        return Candle(value, value, value, value)

    def merge(self, price):
        if isinstance(price, (int, _uint64)):
            return Candle(start=self.start, end=price, min=min(self.min, price), max=max(self.max, price))
        elif isinstance(price, Candle):
            return Candle(start=min(self.start, price.start), end=max(self.start, price.start),
                          min=min(self.min, price.min), max=max(self.max, price.max))
        else:
            raise TypeError("Only integers or candles can be merged to an existing candle")

    def __repr__(self):
        return "Candle(%d, %d, %d, %d)" % (self.start, self.min, self.max, self.end)

    def __str__(self):
        return repr(self)




from datetime import timedelta, date, datetime
from .domains import DiscreteTimeDomain
from .events import Event
from .growing_arrays import GrowingArray


class Timelapse(DiscreteTimeDomain):
    """
    Timelapse is an abstract class allowing us to handle common utilities regarding
      datetimes in frames, digests, and indicators. They will also give yp the data,
      in the end, but will not handle it: they will give the feature to their
      subclasses.
    """

    def __init__(self, dtype, fill_value, chunk_size, width):
        """
        Creates the timelapse.
        :param dtype: The data type.
        :param chunk_size: The chunk size for the underlying growing array.
        :param width: The width of each data item.
        :param fill_value: The value to fill the empty spaces in the data when initialized.
        """

        self._on_refresh_indicators = Event()
        self._data = GrowingArray(dtype, fill_value, chunk_size, width)

    @property
    def on_refresh_indicators(self):
        """
        This event will be triggered on data change so the indicators can update.
        """

        return self._on_refresh_indicators

    @property
    def width(self):
        """
        Returns the width of this timelapse, based on
        what's set in the underlying array's width.
        :return: The width.
        """

        return self._data.width

    def __getitem__(self, item):
        """
        Gets values from the underlying array. It is also allowed to use timestamps instead of
          indices: they will be converted (by truncation/alignment) to the appropriate indices.
        :param item: The item (index or slice) to use to get the data from the underlying array.
        :return:
        """

        if isinstance(item, (date, datetime)):
            item = self.index_for(item)
        elif isinstance(item, slice):
            start = item.start
            stop = item.stop
            if isinstance(start, (date, datetime)):
                start = self.index_for(start)
            if isinstance(stop, (date, datetime)):
                stop = self.index_for(stop)
            item = slice(start, stop, item.step)
        return self._data[item][:]

    def __len__(self):
        return len(self._data)

    @property
    def dtype(self):
        """
        The underlying type of this source frame.
        """

        return self._data.dtype

    def has_item(self, item):
        """
        Tells whether this time slice has populated the given index or stamp, or not yet.
        :param item: The index or stamp to check.
        :return: Whether it has been populated or not.
        """

        if isinstance(item, (date, datetime)):
            item = self.index_for(item)
        return item < len(self._data)



from ...sources import Source


class SidePlucker:
    """
    A side plucker plucks one of the sides of a source flow: the bid or the ask.
    For this to work, the argument for a side plucker must be a source, with any
      type: its type will not care at all.
    """

    BID = 0
    ASK = 1

    pluck_index = None

    def __init__(self, source, side):
        if not isinstance(source, Source):
            raise TypeError("The source must be a Source")
        if side not in (Source.BID, Source.ASK):
            raise ValueError("The side choice must be either Source.BID / 0, or Source.ASK / 1")
        self._source = source
        self._side = side

    @property
    def side(self):
        return self._side

    @property
    def dtype(self):
        return self._source.dtype

    @property
    def initial(self):
        return self._source.initial[self._side]

    def __getitem__(self, item):
        return self._source[item, self._side]

    def __len__(self):
        return len(self._source)




from ...indicators import Indicator
from . import map


class RowPlucker:
    """
    A RowPlucker expects an indicator (which will have dtype
      == float) and a row that must belong to range [0, width),
      and its implementation of __getitem__ wraps the call
      to the wrapped source, and plucks the specified column
      from each row.
    """

    def __init__(self, indicator, row=0):
        if not isinstance(indicator, Indicator) or 0 >= indicator.width():
            raise TypeError("The indicator must be an Indicator, with a well-setup width")
        if not (isinstance(row, int) and 0 <= row < indicator.width()):
            raise ValueError("For an indicator frame, the row argument must be in the range of "
                             "{0 .. width-1} and must be an integer")
        self._row = row
        self._indicator = indicator

    @property
    def row(self):
        return self._row

    def _pluck(self, element):
        return [element[self._row]]

    def __getitem__(self, item):
        return map(self._indicator, item, self._pluck, float)

    def __len__(self):
        return len(self._indicator)



import datetime
import math
import warnings
from enum import IntEnum
from typing import Tuple
import numpy
from . import Indicator
from ..sources import Source
from ..timelapses import Timelapse
from ..utils.mappers.side_pluckers import SidePlucker


class PredictorAlgorithm:
    """
    A predictor algorithm takes an array of elements
    and makes a prediction. It also provides metadata
    of itself that involves interaction with the data
    provided to the indicator.
    """

    def _get_tail_size(self):
        raise NotImplemented

    @property
    def tail_size(self):
        """
        The tail size is: how many elements does this
        predictor instance requires in order to make
        a predictor. If less than those elements are
        provided, then NaN will be the value of both
        the prediction and the structural error of that
        prediction in particular.
        :return: The tail size.
        """

        return self._get_tail_size()

    def _get_step(self):
        raise NotImplemented

    @property
    def step(self):
        """
        The step is: how many steps in the future will this
        predictor instance actually predict. These objects
        consider X to be expressed in units of time, which
        makes the corresponding Y value a function of time,
        which might be a linear or polynomial expression
        or whatever is needed to make a prediction. In this
        case, the step can be freely chosen.

        Unstructured time-series prediction will typically
        have step=1 (constant), while layered time-series
        prediction (where first is the trend, then the season,
        and finally the stationary data) may predict step=N
        while not considering the stationary part important
        enough to suffer when its "long-term" prediction
        converges to a constant.
        """

        return self._get_step()

    def predict(self, x: numpy.ndarray) -> Tuple[float, float]:
        """
        Makes a prediction. The result of the prediction is
        a tuple making: (prediction, structural_error).
        :param x: The input.
        :return: The prediction and the structural error.
        """

        raise NotImplemented


class Predictor(Indicator):
    """
    This is a one-way predictor. Given a series of values, it predicts
    the next value and also provides a bunch of auxiliary values to
    take a look to (e.g. structural coefficient and some notion of MSE
    or related stuff).
    """

    class Columns(IntEnum):
        PREDICTION = 0
        STRUCTURAL_ERROR = 1
        PREDICTION_DIFFERENCE = 2
        STANDARD_ERROR = 3

    def __init__(self, timelapse: Timelapse, algorithm: PredictorAlgorithm,
                 side: int = None, moving_stderr_tail_size: int = 20):
        # First, initialize which data will be read from.
        self._input_data = None
        if isinstance(timelapse, Source):
            if side not in [Source.BID, Source.ASK]:
                raise ValueError("When creating a Predictor indicator from a Source, "
                                 "a side must be chosen and must be either Source.BID "
                                 "or Source.ASK")
            self._input_data = SidePlucker(timelapse, side)
        elif isinstance(timelapse, Indicator):
            if timelapse.width != 1:
                raise ValueError("When creating a Predictor indicator from another indicator, "
                                 "the width of that indicator must be 1. So far, multi-dimensional "
                                 "indicators are not supported yet")
            self._input_data = timelapse
        else:
            raise TypeError("The timelapse must be either a Source or an Indicator")

        # Then, set the predictor instance.
        if not isinstance(algorithm, PredictorAlgorithm) or type(algorithm) == PredictorAlgorithm:
            raise TypeError("The algorithm must be specified and it must be of a strict "
                            "subclass of PredictorAlgorithm")
        self._algorithm = algorithm

        # Finally, the moving STDERR tail size.
        if isinstance(moving_stderr_tail_size, int):
            if moving_stderr_tail_size < 2:
                raise ValueError("The moving standard error tail size must be >= 2")
            if moving_stderr_tail_size < 10:
                warnings.warn("A too small standard deviation tail size was chosen. This will "
                              "work but you might find results less accurate")
        else:
            raise TypeError("The moving standard error tail size must be an integer")
        self._moving_stderr_tail_size = moving_stderr_tail_size
        super().__init__(timelapse)

    def _initial_width(self):
        """
        The initial width for the indicator involves columns:
        - The vector for the prediction.
        - The vector for the structural error for the moment the prediction was done for.
        - The difference between the actual value and the prediction.
        - The standard deviation, taking a proper tail, considering prediction-actual.
        """

        return 4

    def _update(self, start, end):
        """
        Performs a full update, carefully following all the steps.
        :param start: The start position to update.
        :param end: The end (not included) position to update.
        """

        for index in range(start, end):
            self._update_index(index)

    def _update_index(self, index):
        """
        Performs a per-index update, carefully following all the steps.
        :param index: The index being updated.
        """

        # 1. First, take a tail of data. The tail will end
        #    in the given index. If the index is < the tail
        #    size, we'll do nothing at all here.
        if index < self.prediction_tail_size:
            return
        tail = self._input_data[index + 1 - self.prediction_tail_size:index + 1]
        prediction, structural_error = self._algorithm.predict(tail)
        step = self.step
        # 2. Store the prediction in the array (at time {index + step}), at column PREDICTION.
        self._put_value(prediction, index + step, column=self.Columns.PREDICTION)
        # 3. Store the str. error in the array (at time {index + step}), at column STRUCTURAL_ERROR.
        self._put_value(structural_error, index + step, column=self.Columns.STRUCTURAL_ERROR)
        # 4. Store the difference at time {index}, at column PREDICTION_DIFFERENCE. Value:
        #    (self._data[index, PREDICTION] - self._input_data[index]).
        #    It will be NaN if either value is NaN.
        self._put_value(
            self._data[index][self.Columns.PREDICTION] - self._input_data[index],
            index, column=self.Columns.PREDICTION_DIFFERENCE
        )
        # 5. Store the standard error at time {index}, at column STANDARD_ERROR. Value:
        #    if there are at least (moving_stderr_tail_size) elements in the tail:
        #        diffs = self._data[index - moving_stderr_tail_size + 1:index + 1]
        #        variance = (diffs ** 2).sum() / (moving_stderr_tail_size - 1)
        #        self._data[index, STANDARD_ERROR] = sqrt(variance)
        #        if any of these values is NaN, this value will be indeed NaN.
        #    otherwise, let it be NaN as default.
        if index >= self._moving_stderr_tail_size - 1:
            moving_stderr_tail = self._data[
                index - self._moving_stderr_tail_size + 1:index + 1
            ][self.Columns.PREDICTION_DIFFERENCE]
            self._put_value(
                (moving_stderr_tail ** 2).sum() / (self._moving_stderr_tail_size - 1),
                index, column=self.Columns.STANDARD_ERROR
            )

    def _window_from_future(self, item):
        """
        Takes a current window (which can be: an integer, a date,
        or a slice) and gets one that belongs to the future. This
        allows making use of a window that belongs to LATER moments,
        rather than the moments being deemed as "current" in the
        given item.
        :param item: The item.
        :return: The translated (future) item.
        """

        if isinstance(item, int):
            return item + self.step
        if isinstance(item, datetime.datetime):
            return item + datetime.timedelta(seconds=int(self.interval) * self.step)
        if isinstance(item, slice):
            start = item.start
            stop = item.stop
            step = item.step
            return slice(
                self._window_from_future(start),
                self._window_from_future(stop),
                step
            )

    def _get_column_data(self, item, column, when_computed: bool = False):
        """
        Gets the data from a given column for this indicator.
        :param item: The item to get.
        :param column: The column to get the data from.
        :param when_computed: If true, the points will be given
            at the time they were produced, rather than at their
            target time.
        :return: The data.
        """

        return self[self._window_from_future(item) if when_computed else item][:, column]

    def get_prediction(self, item, when_computed: bool = False):
        """
        Gets the prediction for this indicator.
        :param item: The item to get.
        :param when_computed: If true, the points will be given
            at the time they were produced, rather than at their
            prediction time.
        :return: The prediction data.
        """

        return self._get_column_data(item, self.Columns.PREDICTION, when_computed)

    def get_structural_error(self, item, when_computed: bool = False):
        """
        Gets the structural error for this indicator.
        :param item: The item to get.
        :param when_computed: If true, the points will be given
            at the time they were produced, rather than at their
            prediction time.
        :return: The structural error data.
        """

        return self._get_column_data(item, self.Columns.STRUCTURAL_ERROR, when_computed)

    def get_prediction_difference(self, item):
        """
        Gets the prediction difference for this indicator.
        :param item: The item to get.
        :return: The prediction difference data.
        """

        return self._get_column_data(item, self.Columns.PREDICTION_DIFFERENCE)

    def get_trailing_standard_error(self, item):
        """
        Gets the trailing standard error for this indicator.
        :param item: The item to get.
        :return: The trailing standard error data.
        """

        return self._get_column_data(item, self.Columns.STANDARD_ERROR)

    @property
    def prediction_tail_size(self):
        """
        The underlying tail size, according to the algorithm.
        """

        return self._algorithm.tail_size

    @property
    def moving_stderr_tail_size(self):
        """
        The underlying tail size for standard error calculation.
        """

        return self._moving_stderr_tail_size

    @property
    def step(self):
        """
        The distance between the time of the last sample and the
        time, in the future, being predicted.
        """

        return self._algorithm.step



class Event:
    """
    Callbacks can be registered/unregistered to the event,
      and the event can be triggered.
    """

    def __init__(self):
        self._callbacks = set()

    def register(self, callback):
        self._callbacks.add(callback)

    def unregister(self, callback):
        self._callbacks.discard(callback)

    def trigger(self, *args, **kwargs):
        for callback in self._callbacks:
            callback(*args, **kwargs)

    def listeners(self):
        """
        Iterates over all the registered listener methods and their
          bound receivers.
        :return: A generator.
        """
        for callback in self._callbacks:
            yield callback, getattr(callback, '__self__', None)



"""
Providers load or scrap market data to make back-testing or real-time
  population of market data into sources.

Actually, these are two different provider types, and can be compatible
  with certain intervals and data types, while others have different
  compatibilities.
"""

from mistra.core.sources import Source
from .pricing import Candle


class BackTestingProvider:
    """
    Performs a huge bulked data load from a specific historic data source.

    Each back-testing provider will support or allow their own set of intervals,
      and either be via internet or by reading a local file, which may cause an
      heterogeneous range of errors, which are wrapped into a BackTestingProvider.Error
      exception.

    Right now, it is only allowed to create Candle-typed sources, and not price sources
      of StandardizedPrice type.

    To create a provider class, just inherit this class and override `_execute(self)`
      returning a tuple of two results: the buy price source and the sale price source.

    To invoke a provider, just call it like a function:

        source = my_provider_instance()
    """

    class Error(Exception):
        pass

    def _execute(self):
        """
        This abstract method must be implemented to execute the actual provision
          and also ensure to return a source as result, providing both bid/ask
          prices in it..
        """

        raise NotImplemented

    def _merge(self, stamp, buy_price, sale_price, source):
        """
        Merges two prices in certain timestamp into the current source's respective
          candles. If the stamp is not yet populated with the price for that source,
          it makes a constant candle out of it. On the other hand, it merges the input
          price into the current candle.
        :param stamp: The associated timestamp of the prices.
        :param buy_price: The buy price to add or merge, which is standardized (i.e.
          scaled to convert the decimal point to an integer).
        :param sale_price: The sale price to add or merge, which is standardized (i.e.
          scaled to convert the decimal point to an integer).
        :param source: The source to add or merge that price.
        """

        if source.has_item(stamp):
            buy_price = source[stamp][Source.BID].merge(buy_price)
            sale_price = source[stamp][Source.ASK].merge(sale_price)
            source.push((buy_price, sale_price), stamp)
        else:
            source.push((Candle.constant(buy_price),
                         Candle.constant(sale_price)))

    def __call__(self):
        """
        Executes the actual bulk load. Any exception is wrapped into a BackTestingProvider.Error.
        :return: The new source.
        """

        try:
            return self._execute()
        except Exception as e:
            raise self.Error(e)
